from PySide6.QtWidgets import (
    QLineEdit,
    QLabel,
    QSpinBox,
    QHBoxLayout,
    QVBoxLayout,
    QFormLayout,
    QPushButton,
    QFileDialog,
)
from mosamatic2.ui.widgets.panels.visualizations.visualization import Visualization
from mosamatic2.ui.widgets.panels.visualizations.slicevisualization.sliceviewer import SliceViewer
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.ui.settings import Settings
from mosamatic2.ui.utils import is_macos

LOG = LogManager()
PANEL_TITLE = 'SliceVisualization'
PANEL_NAME = 'slicevisualization'


class SliceVisualization(Visualization):
    def __init__(self):
        super(SliceVisualization, self).__init__()
        self.set_title(PANEL_TITLE)
        self._image_line_edit = None
        self._image_select_button = None
        self._image_dir_select_button = None
        self._color_window_spinbox = None
        self._color_level_spinbox = None
        self._color_window_and_level_reset_button = None
        self._load_image_button = None
        self._slice_viewer = None
        self._form_layout = None
        self._settings = None
        self.init_layout()

    def image_line_edit(self):
        if not self._image_line_edit:
            self._image_line_edit = QLineEdit(self.settings().get(f'{PANEL_NAME}/image'))
        return self._image_line_edit
    
    def image_select_button(self):
        if not self._image_select_button:
            self._image_select_button = QPushButton('Select File')
            self._image_select_button.clicked.connect(self.handle_image_select_button)
        return self._image_select_button
    
    def image_dir_select_button(self):
        if not self._image_dir_select_button:
            self._image_dir_select_button = QPushButton('Select Directory')
            self._image_dir_select_button.clicked.connect(self.handle_image_dir_select_button)
        return self._image_dir_select_button
    
    def color_window_spinbox(self):
        if not self._color_window_spinbox:
            self._color_window_spinbox = QSpinBox(self, minimum=0, maximum=1024, value=400)
        return self._color_window_spinbox
    
    def color_level_spinbox(self):
        if not self._color_level_spinbox:
            self._color_level_spinbox = QSpinBox(self, minimum=0, maximum=100, value=40)
        return self._color_level_spinbox
    
    def color_window_and_level_reset_button(self):
        if not self._color_window_and_level_reset_button:
            self._color_window_and_level_reset_button = QPushButton('Reset')
            self._color_window_and_level_reset_button.clicked.connect(self.handle_color_window_and_level_reset_button)
        return self._color_window_and_level_reset_button
    
    def load_image_button(self):
        if not self._load_image_button:
            self._load_image_button = QPushButton('Load')
            self._load_image_button.clicked.connect(self.handle_load_image_button)
        return self._load_image_button
    
    def slice_viewer(self):
        if not self._slice_viewer:
            self._slice_viewer = SliceViewer(
                color_window=self.color_window_spinbox().value(),
                color_level=self.color_level_spinbox().value(),
            )
        return self._slice_viewer

    def form_layout(self):
        if not self._form_layout:
            self._form_layout = QFormLayout()
            if is_macos():
                self._form_layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)
        return self._form_layout

    def settings(self):
        if not self._settings:
            self._settings = Settings()
        return self._settings

    def init_layout(self):
        image_layout = QHBoxLayout()
        image_layout.addWidget(self.image_line_edit())
        image_layout.addWidget(self.image_select_button())
        image_layout.addWidget(self.image_dir_select_button())
        color_window_and_level_layout = QHBoxLayout()
        color_window_and_level_layout.addWidget(QLabel('Color window'))
        color_window_and_level_layout.addWidget(self.color_window_spinbox())
        color_window_and_level_layout.addWidget(QLabel('Color level'))
        color_window_and_level_layout.addWidget(self.color_level_spinbox())
        color_window_and_level_layout.addWidget(self.color_window_and_level_reset_button())
        self.form_layout().addRow('NIFTI file or DICOM directory', image_layout)
        self.form_layout().addRow('', color_window_and_level_layout)
        layout = QVBoxLayout()
        layout.addLayout(self.form_layout())
        layout.addWidget(self.load_image_button())
        layout.addWidget(self.slice_viewer())
        self.setLayout(layout)
        self.setObjectName(PANEL_NAME)

    def handle_image_select_button(self):
        last_directory = self.settings().get('last_directory')
        file_path, _ = QFileDialog.getOpenFileName(dir=last_directory)
        if file_path:
            self.image_line_edit().setText(file_path)
            self.settings().set('last_directory', file_path)

    def handle_image_dir_select_button(self):
        last_directory = self.settings().get('last_directory')
        dir_path = QFileDialog.getExistingDirectory(dir=last_directory)
        if dir_path:
            self.image_line_edit().setText(dir_path)
            self.settings().set('last_directory', dir_path)

    def handle_color_window_and_level_reset_button(self):
        self.slice_viewer().set_color_window(self.color_window_spinbox().value())
        self.slice_viewer().set_color_level(self.color_level_spinbox().value())

    def handle_load_image_button(self):
        self.slice_viewer().set_nifti_file_or_dicom_dir(self.image_line_edit().text())
        # self.slice_viewer().set_view_orientation('axial')
        self.slice_viewer().load_image()

    def save_inputs_and_parameters(self):
        self.settings().set(f'{PANEL_NAME}/image', self.image_line_edit().text())


import os
import zipfile
import tempfile
import numpy as np

import models

from mosamatic2.core.tasks.task import Task
from mosamatic2.core.tasks.segmentmusclefatl3tensorflowtask.paramloader import ParamLoader
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.data.dicomimage import DicomImage
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import (
    normalize_between,
    get_pixels_from_dicom_object,
    convert_labels_to_157,
)

DEVICE = 'cpu'
L3_INDEX = 167
LOG = LogManager()


class SegmentMuscleFatL3TensorFlowTask(Task):
    INPUTS = [
        'images', 
        'model_files'
    ]
    PARAMS = ['model_version']

    def __init__(self, inputs, params, output, overwrite=True):
        super(SegmentMuscleFatL3TensorFlowTask, self).__init__(inputs, params, output, overwrite)

    def load_images(self):        
        image_data = MultiDicomImage()
        image_data.set_path(self.input('images'))
        if image_data.load():
            return image_data
        raise RuntimeError('Could not load images')

    def load_model_files(self):
        model_files = []
        for f in os.listdir(self.input('model_files')):
            f_path = os.path.join(self.input('model_files'), f)
            if f_path.endswith('.zip') or f_path.endswith('.json'):
                model_files.append(f_path)
        if len(model_files) != 3:
            raise RuntimeError(f'Found {len(model_files)} model files. This should be 3!')
        return model_files

    def load_models_and_params(self, model_files, model_version):
        tfLoaded = False
        model, contour_model, params = None, None, None
        for f_path in model_files:
            f_name = os.path.split(f_path)[1]
            if f_name == f'model-{str(model_version)}.zip':
                if not tfLoaded:
                    import tensorflow as tf
                    tfLoaded = True
                with tempfile.TemporaryDirectory() as model_dir_unzipped:
                    os.makedirs(model_dir_unzipped, exist_ok=True)
                    with zipfile.ZipFile(f_path) as zipObj:
                        zipObj.extractall(path=model_dir_unzipped)
                    model = tf.keras.models.load_model(model_dir_unzipped, compile=False)
            elif f_name == f'contour_model-{str(model_version)}.zip':
                if not tfLoaded:
                    import tensorflow as tf
                    tfLoaded = True
                with tempfile.TemporaryDirectory() as contour_model_dir_unzipped:
                    os.makedirs(contour_model_dir_unzipped, exist_ok=True)
                    with zipfile.ZipFile(f_path) as zipObj:
                        zipObj.extractall(path=contour_model_dir_unzipped)
                    contour_model = tf.keras.models.load_model(contour_model_dir_unzipped, compile=False)
            elif f_name == f'params-{model_version}.json':
                params = ParamLoader(f_path)
            else:
                pass
        return model, contour_model, params

    def extract_contour(self, image, contour_model, params):
        ct = np.copy(image)
        ct = normalize_between(ct, params.dict['min_bound_contour'], params.dict['max_bound_contour'])
        img2 = np.expand_dims(ct, 0)
        img2 = np.expand_dims(img2, -1)
        pred = contour_model.predict([img2])
        pred_squeeze = np.squeeze(pred)
        pred_max = pred_squeeze.argmax(axis=-1)
        mask = np.uint8(pred_max)
        return mask

    def segment_muscle_and_fat(self, image, model):
        img2 = np.expand_dims(image, 0)
        img2 = np.expand_dims(img2, -1)
        pred = model.predict([img2])
        pred_squeeze = np.squeeze(pred)
        pred_max = pred_squeeze.argmax(axis=-1)
        return pred_max
        
    def process_file(self, image, output_dir, model, contour_model, params):
        assert isinstance(image, DicomImage)
        pixels = get_pixels_from_dicom_object(image.object(), normalize=True)
        if contour_model:
            mask = self.extract_contour(pixels, contour_model, params)
            pixels = normalize_between(pixels, params.dict['min_bound'], params.dict['max_bound'])
            pixels = pixels * mask
        pixels = pixels.astype(np.float32)
        segmentation = self.segment_muscle_and_fat(pixels, model)
        segmentation = convert_labels_to_157(segmentation)
        segmentation_file_name = os.path.split(image.path())[1]
        segmentation_file_path = os.path.join(output_dir, f'{segmentation_file_name}.seg.npy')
        np.save(segmentation_file_path, segmentation)

    def run(self):
        image_data = self.load_images()
        model_files = self.load_model_files()
        model_version = self.param('model_version')
        model, contour_model, params = self.load_models_and_params(model_files, model_version)
        images = image_data.images()
        nr_steps = len(images)
        for step in range(nr_steps):
            self.process_file(images[step], self.output(), model, contour_model, params)
            self.set_progress(step, nr_steps)


import os
import numpy as np
import pandas as pd
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.data.numpyimage import NumpyImage
from mosamatic2.core.utils import (
    get_pixels_from_dicom_object,
    calculate_area,
    calculate_mean_radiation_attenuation,
    calculate_lama_percentage,
    get_pixels_from_tag_file,
    MUSCLE,
    SAT,
    VAT,
)

LOG = LogManager()


class CalculateScoresTask(Task):
    INPUTS = [
        'images', 
        'segmentations'
    ]
    PARAMS = ['file_type']

    def __init__(self, inputs, params, output, overwrite=True):
        super(CalculateScoresTask, self).__init__(inputs, params, output, overwrite)

    def collect_img_seg_pairs(self, images, segmentations, file_type='npy'):
        file_type = '.tag' if file_type == 'tag' else '.seg.npy'
        img_seg_pairs = []
        for image in images:
            f_img_path = image.path()
            f_img_name = os.path.split(f_img_path)[1]
            for f_seg_path in segmentations:
                f_seg_name = os.path.split(f_seg_path)[1]
                if file_type == '.seg.npy':
                    f_seg_name = f_seg_name.removesuffix(file_type)
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                elif file_type == '.tag':
                    f_seg_name = f_seg_name.removesuffix(file_type).removesuffix('.dcm')
                    f_img_name = f_img_name.removesuffix('.dcm')
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                else:
                    raise RuntimeError('Unknown file type')
        return img_seg_pairs

    def load_images(self):
        image_data = MultiDicomImage()
        image_data.set_path(self.input('images'))
        if image_data.load():
            return image_data
        raise RuntimeError('Could not load images')
    
    def load_pixels_and_spacing(self, image):
        p = image.object()
        pixels = get_pixels_from_dicom_object(p, normalize=True)
        return pixels, p.PixelSpacing

    def load_segmentations(self, file_type='npy'):
        file_type = '.tag' if file_type == 'tag' else '.seg.npy'
        segmentations = []
        for f in os.listdir(self.input('segmentations')):
            f_path = os.path.join(self.input('segmentations'), f)
            if f.endswith(file_type):
                segmentations.append(f_path)
        if len(segmentations) == 0:
            raise RuntimeError('Input directory has no segmentation files')
        return segmentations

    def load_segmentation(self, f, file_type='npy'):
        if file_type == 'npy':
            segmentation = NumpyImage()
            segmentation.set_path(f)
            if segmentation.load():
                return segmentation.object()
            LOG.error(f'Could not load segmentation file {f}')
        if file_type == 'tag':
            pixels = get_pixels_from_tag_file(f)
            try:
                pixels = pixels.reshape(512, 512)
                return pixels
            except Exception:
                LOG.warning(f'Could not reshape TAG pixels to (512, 512), skipping...')
                return None
        raise RuntimeError('Unknown file type')

    def run(self):
        image_data = self.load_images()
        images = image_data.images()
        file_type = self.param('file_type')
        segmentations = self.load_segmentations(file_type)
        img_seg_pairs = self.collect_img_seg_pairs(images, segmentations, file_type)
        # Create empty data dictionary
        data = {
            'file': [], 
            'muscle_area': [], 'muscle_idx': [], 'muscle_ra': [], 'muscle_lama_perc': [],
            'vat_area': [], 'vat_idx': [], 'vat_ra': [],
            'sat_area': [], 'sat_idx': [], 'sat_ra': []
        }
        nr_steps = len(img_seg_pairs)
        for step in range(nr_steps):
            # Get image and its pixel spacing
            image, pixel_spacing = self.load_pixels_and_spacing(img_seg_pairs[step][0])
            if image is None:
                raise RuntimeError(f'Could not load DICOM image for file {img_seg_pairs[step][0]}')
            # Get segmentation for this image
            segmentation = self.load_segmentation(img_seg_pairs[step][1], file_type)
            if segmentation is None:
                LOG.warning(f'Could not load segmentation for file {img_seg_pairs[step][1]}')
                continue
            # Calculate metrics
            file_name = os.path.split(img_seg_pairs[step][0].path())[1]
            muscle_area = calculate_area(segmentation, MUSCLE, pixel_spacing)
            muscle_idx = 0
            muscle_ra = calculate_mean_radiation_attenuation(image, segmentation, MUSCLE)
            muscle_lama_perc = calculate_lama_percentage(image, segmentation, MUSCLE)
            vat_area = calculate_area(segmentation, VAT, pixel_spacing)
            vat_idx = 0
            vat_ra = calculate_mean_radiation_attenuation(image, segmentation, VAT)
            sat_area = calculate_area(segmentation, SAT, pixel_spacing)
            sat_idx = 0
            sat_ra = calculate_mean_radiation_attenuation(image, segmentation, SAT)
            LOG.info(f'file: {file_name}, ' +
                    f'muscle_area: {muscle_area}, muscle_idx: {muscle_idx}, muscle_ra: {muscle_ra}, ' +
                    f'vat_area: {vat_area}, vat_idx: {vat_idx}, vat_ra: {vat_ra}, ' +
                    f'sat_area: {sat_area}, sat_idx: {sat_idx}, sat_ra: {sat_ra}')
            # Update dataframe data
            data['file'].append(file_name)
            data['muscle_area'].append(muscle_area)
            data['muscle_idx'].append(muscle_idx)
            data['muscle_ra'].append(muscle_ra)
            data['muscle_lama_perc'].append(muscle_lama_perc)
            data['vat_area'].append(vat_area)
            data['vat_idx'].append(vat_idx)
            data['vat_ra'].append(vat_ra)
            data['sat_area'].append(sat_area)
            data['sat_idx'].append(sat_idx)
            data['sat_ra'].append(sat_ra)
            # Update progress
            self.set_progress(step, nr_steps)
        # Build dataframe and return the CSV file as output
        csv_file_path = os.path.join(self.output(), 'bc_scores.csv')
        xls_file_path = os.path.join(self.output(), 'bc_scores.xlsx')
        df = pd.DataFrame(data=data)
        df.to_csv(csv_file_path, index=False, sep=';')
        df.to_excel(xls_file_path, index=False, engine='openpyxl')


import os
import json
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import is_dicom, load_dicom

LOG = LogManager()


class CreateDicomSummaryTask(Task):
    INPUTS = ['directory']
    PARAMS = []

    def __init__(self, inputs, params, output, overwrite):
        super(CreateDicomSummaryTask, self).__init__(inputs, params, output, overwrite)

    def run(self):
        nr_steps = len(os.listdir(self.input('directory'))) * 2
        LOG.info('Collecting data...')
        data = {}
        step = 0
        for patient_dir_name in os.listdir(self.input('directory')):
            data[patient_dir_name] = {}
            patient_dir_path = os.path.join(self.input('directory'), patient_dir_name)
            for root, dirs, files in os.walk(patient_dir_path):
                for f in files:
                    f_path = os.path.join(root, f)
                    # if is_dicom(f_path):
                    p = load_dicom(f_path, stop_before_pixels=True)
                    if p:
                        series_instance_uid = p.SeriesInstanceUID
                        if not series_instance_uid in data[patient_dir_name].keys():
                            data[patient_dir_name][series_instance_uid] = []
                        data[patient_dir_name][series_instance_uid].append(p)
            self.set_progress(step, nr_steps)
            step += 1
        LOG.info('Building summary...')
        summary = {
            'nr_series_per_patient': {},
        }
        # Calculate nr. of series per patient
        for patient_dir_name in data.keys():
            summary['nr_series_per_patient'][patient_dir_name] = len(data[patient_dir_name].keys())
        # Summarize series data per patient
        for patient_dir_name in data.keys():
            summary[patient_dir_name] = {}
            for series_instance_uid in data[patient_dir_name].keys():
                summary[patient_dir_name][series_instance_uid] = {
                    'nr_images': len(data[patient_dir_name][series_instance_uid]),
                    'modality': data[patient_dir_name][series_instance_uid][0].Modality,
                    'image_type': str(data[patient_dir_name][series_instance_uid][0].ImageType),
                    'rows': data[patient_dir_name][series_instance_uid][0].Rows,
                    'columns': data[patient_dir_name][series_instance_uid][0].Columns,
                    'pixel_spacing': str(data[patient_dir_name][series_instance_uid][0].PixelSpacing),
                    'slice_thickness': data[patient_dir_name][series_instance_uid][0].SliceThickness,
                }
            self.set_progress(step, nr_steps)
            step += 1
        LOG.info('Exporting summary...')
        with open(os.path.join(self.output(), 'summary.txt'), 'w') as f:
            json.dump(summary, f, indent=4)


import click

from mosamatic2.core.tasks import SelectSliceFromScansTask


@click.command(help='Selects specific slice from CT scans')
@click.option(
    '--scans', 
    required=True, 
    type=click.Path(exists=True), 
    help='Directory with scans (each patient separate subdirectory)',
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory'
)
@click.option(
    '--vertebra', 
    required=True,
    default='L3',
    help='Vertebral level for selecting slice (default: "L3")'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=True, 
    help='Overwrite [true|false] (default: true)'
)
def selectslicefromscans(scans, vertebra, output, overwrite):
    """
    Selects specific slice from list of CT scans
    
    Parameters
    ----------
    --scans : str
        Directory to scans. Each patient's scan should be in a separate
        subdirectory. For example:

        /scans
        /scans/patient1
        /scans/patient1/file1.dcm
        /scans/patient1/file2.dcm
        ...
        /scans/patient2
        ...

    --output : str
        Path to output directory where selected slices will be placed. Each
        slice's file name will be the same as the scan directory name, so in
        the example above that would be "patient1", "patient2", etc.

    --vertebra : str
        Vertebral level where to take slice [L3|T4] (default: L3)

    --overwrite : bool
        Overwrite contents output directory [true|false] (default: true)
    """
    task = SelectSliceFromScansTask(
        inputs={'scans': scans},
        params={'vertebra': vertebra},
        output=output,
        overwrite=overwrite,
    )
    task.run()


import click

from mosamatic2.core.tasks import CreateDicomSummaryTask


@click.command(help='Creates a DICOM summary inside a root directory')
@click.option(
    '--directory', 
    required=True, 
    type=click.Path(exists=True), 
    help='Root directory with DICOM images (can be multiple scans)',
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def createdicomsummary(directory, output, overwrite):
    """
    Creates a DICOM summary inside a root directory. Each patient should have
    its own directory. Inside each patient's directory, there can be multiple
    scans (called series) and multiple DICOM files per scan. The output of this
    command is a file summary.txt (stored in the output directory) that contains
    the following information:

    - A list of patient directory names with the number of scans inside each
      patient directory
    - For each patient directory and scan directory the following information:
        - Nr. of DICOM images in the scan
        - Modality (e.g., CT or MRI)
        - Image type (e.g., for Dixon scans can be in-phase, opposite-phase, 
          water or fat images)
        - Rows/columns (size of the images)
        - Pixel spacing (size of each pixel in mm^2)
        - Slice thickness: (thickness of each image slice)

    Parameters
    ----------
    --directory : str
        Root directory with patient directories, scans and DICOM images

    --output : str
        Path to output directory

    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    task = CreateDicomSummaryTask(
        inputs={'directory': directory},
        params=None,
        output=output,
        overwrite=overwrite,
    )
    task.run()


import json


class ParamLoader:
    def __init__(self, json_path):
        self.update(json_path)

    def save(self, json_path):
        """"
        Save dict to json file

        Parameters
        ----------
        json_path : string
            Path to save location
        """
        with open(json_path, 'w') as f:
            json.dump(self.__dict__, f, indent=4)

    def update(self, json_path):
        """
        Load parameters from json file

        Parameters
        ----------
        json_path : string
            Path to json file
        """
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    @property
    def dict(self):
        """"
        Give dict-like access to Params instance
        by: 'params.dict['learning_rate']'
        """
        return self.__dict__



from mosamatic2.core.tasks.rescaledicomimagestask.rescaledicomimagestask import RescaleDicomImagesTask
from mosamatic2.core.tasks.segmentmusclefatl3tensorflowtask.segmentmusclefatl3tensorflowtask import SegmentMuscleFatL3TensorFlowTask
from mosamatic2.core.tasks.calculatescorestask.calculatescorestask import CalculateScoresTask
from mosamatic2.core.tasks.createpngsfromsegmentationstask.createpngsfromsegmentationstask import CreatePngsFromSegmentationsTask
from mosamatic2.core.tasks.dicom2niftitask.dicom2niftitask import Dicom2NiftiTask
from mosamatic2.core.tasks.selectslicefromscanstask.selectslicefromscanstask import SelectSliceFromScansTask
from mosamatic2.core.tasks.createdicomsummarytask.createdicomsummarytask import CreateDicomSummaryTask
from mosamatic2.core.tasks.totalsegmentatortask.totalsegmentatortask import TotalSegmentatorTask
from mosamatic2.core.tasks.calculatemaskstatisticstask.calculatemaskstatisticstask import CalculateMaskStatisticsTask
from mosamatic2.core.tasks.segmentmusclefatpytorchtask.segmentmusclefatpytorchtask import SegmentMuscleFatPyTorchTask


import os
from mosamatic2.core.pipelines.pipeline import Pipeline
from mosamatic2.core.tasks import (
    Dicom2NiftiTask,
    TotalSegmentatorTask,
    CalculateMaskStatisticsTask,
)
from mosamatic2.core.managers.logmanager import LogManager

LOG = LogManager()


class LiverAnalysisPipeline(Pipeline):
    INPUTS = ['scans']
    PARAMS = ['compressed']

    def __init__(self, inputs, params, output, overwrite):
        super(LiverAnalysisPipeline, self).__init__(inputs, params, output, overwrite)
        self.add_task(
            Dicom2NiftiTask(
                inputs={'scans': self.input('scans')},
                params={'compressed': self.param('compressed')},
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )
        self.add_task(
            TotalSegmentatorTask(
                inputs={'scans': os.path.join(self.output(), 'dicom2niftitask')},
                params={
                    'tasks': 'liver_segments', 
                    'format': 'nifti',
                },
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )
        self.add_task(
            CalculateMaskStatisticsTask(
                inputs={
                    'scans': os.path.join(self.output(), 'dicom2niftitask'),
                    'masks': os.path.join(self.output(), 'totalsegmentatortask'),
                },
                params=None,
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )


from mosamatic2.core.data.filedata import FileData


class DixonSeries(FileData):
    def __init__(self):
        super(DixonSeries, self).__init__()
        self._series = {'ip': None, 'op': None, 'water': None, 'fat': None}

    def ip(self):
        return self._series['ip']
    
    def op(self):
        return self._series['op']
    
    def water(self):
        return self._series['water']
    
    def fat(self):
        return self._series['fat']
    
    def load(self):
        pass


import click

from mosamatic2.core.pipelines import BoaDockerPipeline


@click.command(help='Runs 3D body composition analysis pipeline using BOA algorithm')
@click.option(
    '--scans', 
    required=True, 
    type=click.Path(exists=True), 
    help='Root directory with scan directories for each patient (no spaces allowed)',
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory (no spaces allowed)'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def boadockerpipeline(scans, output, overwrite):
    """
    Runs 3D body composition analysis pipeline using BOA algorithm.
    
    Parameters
    ----------
    --scans : str
        Root directory with scan directories for each patient. Each scan directory should 
        contain DICOM images for a single scan only and nothing else. Also, the directory
        paths cannot contain any spaces.

    --output : str
        Path to output directory. No spaces allowed.

    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    pipeline = BoaDockerPipeline(
        inputs={'scans': scans},
        params=None,
        output=output,
        overwrite=overwrite,
    )
    pipeline.run()


import click

from mosamatic2.core.tasks import CreatePngsFromSegmentationsTask


@click.command(help='Create PNG images from segmentation files')
@click.option(
    '--segmentations', 
    required=True, 
    type=click.Path(exists=True), 
    help='Input directory to segmentation files'
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory for PNG images'
)
@click.option(
    '--fig_width', 
    type=click.Path(), 
    default=10,
    help='Figure width (default: 10)'
)
@click.option(
    '--fig_height', 
    type=click.Path(), 
    default=10,
    help='Figure height (default: 10)'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def createpngsfromsegmentations(segmentations, output, fig_width, fig_height, overwrite):
    """
    Creates PNG images from L3 muscle and fat segmentation files
    
    Parameters
    ----------
    --segmentations : str
        Directory with with input segmentation files. Must be L3 muscle and fat segmentations
        (output of "mosamatic2-cli segmentmusclefatl3tensorflow")

    --output : str
        Path to output directory

    --fig_width : int
        Width of PNG image

    --fig_height : int
        Height of PNG image

    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    task = CreatePngsFromSegmentationsTask(
        inputs={'segmentations': segmentations},
        params={'fig_width': fig_width, 'fig_height': fig_height},
        output=output,
        overwrite=overwrite,
    )
    task.run()


import os
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.data.filedata import FileData
from mosamatic2.core.data.dicomimage import DicomImage

LOG = LogManager()


class MultiDicomImage(FileData):
    def __init__(self):
        super(MultiDicomImage, self).__init__()
        self._images = []

    def images(self):
        return self._images
    
    def load(self):
        series_instance_uids = []
        if self.path():
            for f in os.listdir(self.path()):
                f_path = os.path.join(self.path(), f)
                image = DicomImage()
                image.set_path(f_path)
                if image.load():
                    series_instance_uid = image.object().SeriesInstanceUID
                    if series_instance_uid in series_instance_uids:
                        RuntimeError('Cannot load DICOM images with identical series instance UID')
                    self._images.append(image)
            return True
        return False


import os
import mosamatic2.constants as constants
from PySide6.QtWidgets import (
    QMainWindow,
)
from PySide6.QtGui import (
    QGuiApplication,
    QAction,
    QIcon,
)
from PySide6.QtCore import Qt, QByteArray
from mosamatic2.ui.utils import version, resource_path, is_macos
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.ui.settings import Settings
from mosamatic2.ui.widgets.panels.mainpanel import MainPanel
from mosamatic2.ui.widgets.panels.logpanel import LogPanel
from mosamatic2.ui.widgets.panels.tasks.rescaledicomimagestaskpanel import RescaleDicomImagesTaskPanel
from mosamatic2.ui.widgets.panels.tasks.segmentmusclefatl3tensorflowtaskpanel import SegmentMuscleFatL3TensorFlowTaskPanel
from mosamatic2.ui.widgets.panels.tasks.segmentmusclefatpytorchtaskpanel import SegmentMuscleFatPyTorchTaskPanel
from mosamatic2.ui.widgets.panels.tasks.createpngsfromsegmentationstaskpanel import CreatePngsFromSegmentationsTaskPanel
from mosamatic2.ui.widgets.panels.tasks.calculatescorestaskpanel import CalculateScoresTaskPanel
from mosamatic2.ui.widgets.panels.tasks.dicom2niftitaskpanel import Dicom2NiftiTaskPanel
from mosamatic2.ui.widgets.panels.tasks.createdicomsummarytaskpanel import CreateDicomSummaryTaskPanel
from mosamatic2.ui.widgets.panels.tasks.selectslicefromscanstaskpanel import SelectSliceFromScansTaskPanel
from mosamatic2.ui.widgets.panels.tasks.totalsegmentatortaskpanel import TotalSegmentatorTaskPanel
from mosamatic2.ui.widgets.panels.tasks.calculatemaskstatisticstaskpanel import CalculateMaskStatisticsTaskPanel
from mosamatic2.ui.widgets.panels.pipelines.defaultpipelinepanel import DefaultPipelinePanel
from mosamatic2.ui.widgets.panels.pipelines.defaultdockerpipelinepanel import DefaultDockerPipelinePanel
from mosamatic2.ui.widgets.panels.pipelines.boadockerpipelinepanel import BoaDockerPipelinePanel
from mosamatic2.ui.widgets.panels.pipelines.liveranalysispipelinepanel import LiverAnalysisPipelinePanel
from mosamatic2.ui.widgets.panels.visualizations.slicevisualization.slicevisualization import SliceVisualization

LOG = LogManager()


class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self._settings = None
        self._main_panel = None
        self._log_panel = None
        self._decompress_dicom_files_task_panel = None
        self._rescale_dicom_images_task_panel = None
        self._segment_muscle_fat_l3_tensorflow_task_panel = None
        self._segment_muscle_fat_pytorch_task_panel = None
        self._create_pngs_from_segmentations_task_panel = None
        self._calculate_scores_task_panel = None
        self._dicom2nifti_task_panel = None
        self._create_dicom_summary_task_panel = None
        self._select_slice_from_scans_task_panel = None
        self._total_segmentator_task_panel = None
        self._calculate_mask_statistics_task_panel = None
        self._default_pipeline_panel = None
        self._default_docker_pipeline_panel = None
        self._boa_docker_pipeline_panel = None
        self._liver_analysis_pipeline_panel = None
        self._slice_visualization = None
        self.init_window()

    def init_window(self):
        self.setWindowTitle(f'{constants.MOSAMATIC2_WINDOW_TITLE} {version()}')
        icon_file_name = constants.MOSAMATIC2_APP_ICON_FILE_NAME_MAC if is_macos() else constants.MOSAMATIC2_APP_ICON_FILE_NAME_WIN
        icon_path = resource_path(os.path.join(constants.MOSAMATIC2_ICONS_DIR_PATH, icon_file_name))
        self.setWindowIcon(QIcon(icon_path))
        if not self.load_geometry_and_state():
            self.set_default_size_and_position()
        self.init_menus()
        self.init_status_bar()
        self.addDockWidget(Qt.DockWidgetArea.TopDockWidgetArea, self.main_panel())
        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.log_panel())

    def init_menus(self):
        self.init_app_menu()
        self.init_tasks_menu()
        self.init_pipelines_menu()
        self.init_visualizations_menu()
        if is_macos():            
            self.menuBar().setNativeMenuBar(False)

    def init_app_menu(self):
        exit_action = QAction('Exit', self)
        exit_action.triggered.connect(self.close)
        app_menu = self.menuBar().addMenu('Application')
        app_menu.addAction(exit_action)

    def init_tasks_menu(self):
        rescale_dicom_images_task_action = QAction('RescaleDicomImagesTask', self)
        rescale_dicom_images_task_action.triggered.connect(self.handle_rescale_dicom_images_task_action)
        segment_muscle_fat_l3_tensorflow_task_action = QAction('SegmentMuscleAndFatTensorFlowTask', self)
        segment_muscle_fat_l3_tensorflow_task_action.triggered.connect(self.handle_segment_muscle_fat_l3_tensorflow_task_action)
        segment_muscle_fat_pytorch_task_action = QAction("SegmentMuscleFatPytorchTask", self)
        segment_muscle_fat_pytorch_task_action.triggered.connect(self.handle_segment_muscle_fat_pytorch_task_action)
        calculate_scores_task_action = QAction('CalculateScoresTask', self)
        calculate_scores_task_action.triggered.connect(self.handle_calculate_scores_task_action)
        create_pngs_from_segmentations_task_action = QAction('CreatePngsFromSegmentationsTask', self)
        create_pngs_from_segmentations_task_action.triggered.connect(self.handle_create_pngs_from_segmentations_task_action)
        dicom2nifti_task_action = QAction('Dicom2NiftiTask', self)
        dicom2nifti_task_action.triggered.connect(self.handle_dicom2nifti_task_action)
        create_dicom_summary_task_action = QAction('CreateDicomSummaryTask', self)
        create_dicom_summary_task_action.triggered.connect(self.handle_create_dicom_summary_task_action)
        select_slice_from_scans_task_action = QAction('SelectSliceFromScansTask', self)
        select_slice_from_scans_task_action.triggered.connect(self.handle_select_slice_from_scans_task_action)
        total_segmentator_task_action = QAction('TotalSegmentatorTask', self)
        total_segmentator_task_action.triggered.connect(self.handle_total_segmentator_task_action)
        calculate_mask_statistics_task_action = QAction('CalculateMaskStatisticsTask', self)
        calculate_mask_statistics_task_action.triggered.connect(self.handle_calculate_mask_statistics_task_action)
        tasks_menu = self.menuBar().addMenu('Tasks')
        tasks_menu.addAction(rescale_dicom_images_task_action)
        tasks_menu.addAction(segment_muscle_fat_l3_tensorflow_task_action)
        tasks_menu.addAction(segment_muscle_fat_pytorch_task_action)
        tasks_menu.addAction(calculate_scores_task_action)
        tasks_menu.addAction(create_pngs_from_segmentations_task_action)
        tasks_menu.addAction(dicom2nifti_task_action)
        tasks_menu.addAction(create_dicom_summary_task_action)
        tasks_menu.addAction(select_slice_from_scans_task_action)
        tasks_menu.addAction(total_segmentator_task_action)
        tasks_menu.addAction(calculate_mask_statistics_task_action)

    def init_pipelines_menu(self):
        default_pipeline_action = QAction('DefaultPipeline', self)
        default_pipeline_action.triggered.connect(self.handle_default_pipeline_action)
        default_docker_pipeline_action = QAction('DefaultDockerPipeline', self)
        default_docker_pipeline_action.triggered.connect(self.handle_default_docker_pipeline_action)
        boa_docker_pipeline_action = QAction('BoaDockerPipeline', self)
        boa_docker_pipeline_action.triggered.connect(self.handle_boa_docker_pipeline_action)
        liver_analysis_pipeline_action = QAction('LiverAnalysisPipeline', self)
        liver_analysis_pipeline_action.triggered.connect(self.handle_liver_analysis_pipeline_action)
        pipelines_menu = self.menuBar().addMenu('Pipelines')
        pipelines_menu.addAction(default_pipeline_action)
        pipelines_menu.addAction(default_docker_pipeline_action)
        pipelines_menu.addAction(boa_docker_pipeline_action)
        pipelines_menu.addAction(liver_analysis_pipeline_action)

    def init_visualizations_menu(self):
        slice_visualization_action = QAction('SliceVisualization', self)
        slice_visualization_action.triggered.connect(self.handle_slice_visualization_action)
        visualizations_menu = self.menuBar().addMenu('Visualizations')
        visualizations_menu.addAction(slice_visualization_action)

    def init_status_bar(self):
        self.set_status('Ready')

    # GET

    def settings(self):
        if not self._settings:
            self._settings = Settings()
        return self._settings

    def main_panel(self):
        if not self._main_panel:
            self._main_panel = MainPanel(self)
            self._main_panel.add_panel(self.rescale_dicom_images_task_panel(), 'rescaledicomimagestaskpanel')
            self._main_panel.add_panel(self.segment_muscle_fat_l3_tensorflow_task_panel(), 'segmentmusclefatl3tensorflowtaskpanel')
            self._main_panel.add_panel(self.segment_muscle_fat_pytorch_task_panel(), 'segmentmusclefatpytorchtaskpanel')
            self._main_panel.add_panel(self.create_pngs_from_segmentations_task_panel(), 'createpngsfromsegmentationstaskpanel')
            self._main_panel.add_panel(self.calculate_scores_task_panel(), 'calculatescorestaskpanel')
            self._main_panel.add_panel(self.dicom2nifti_task_panel(), 'dicom2niftitaskpanel')
            self._main_panel.add_panel(self.create_dicom_summary_task_panel(), 'createdicomsummarytaskpanel')
            self._main_panel.add_panel(self.select_slice_from_scans_task_panel(), 'selectslicefromscanstaskpanel')
            self._main_panel.add_panel(self.total_segmentator_task_panel(), 'totalsegmentatortaskpanel')
            self._main_panel.add_panel(self.calculate_mask_statistics_task_panel(), 'calculatemaskstatisticstaskpanel')
            self._main_panel.add_panel(self.default_pipeline_panel(), 'defaultpipelinepanel')
            self._main_panel.add_panel(self.default_docker_pipeline_panel(), 'defaultdockerpipelinepanel')
            self._main_panel.add_panel(self.boa_docker_pipeline_panel(), 'boadockerpipelinepanel')
            self._main_panel.add_panel(self.liver_analysis_pipeline_panel(), 'liveranalysispipelinepanel')
            self._main_panel.add_panel(self.slice_visualization(), 'slicevisualization')
            self._main_panel.select_panel('defaultpipelinepanel')
        return self._main_panel
    
    def log_panel(self):
        if not self._log_panel:
            self._log_panel = LogPanel()
            LOG.add_listener(self._log_panel)
        return self._log_panel

    # MISCELLANEOUS

    def rescale_dicom_images_task_panel(self):
        if not self._rescale_dicom_images_task_panel:
            self._rescale_dicom_images_task_panel = RescaleDicomImagesTaskPanel()
        return self._rescale_dicom_images_task_panel
    
    def segment_muscle_fat_l3_tensorflow_task_panel(self):
        if not self._segment_muscle_fat_l3_tensorflow_task_panel:
            self._segment_muscle_fat_l3_tensorflow_task_panel = SegmentMuscleFatL3TensorFlowTaskPanel()
        return self._segment_muscle_fat_l3_tensorflow_task_panel

    def segment_muscle_fat_pytorch_task_panel(self):
        if not self._segment_muscle_fat_pytorch_task_panel:
            self._segment_muscle_fat_pytorch_task_panel = SegmentMuscleFatPyTorchTaskPanel()
        return self._segment_muscle_fat_pytorch_task_panel

    def create_pngs_from_segmentations_task_panel(self):
        if not self._create_pngs_from_segmentations_task_panel:
            self._create_pngs_from_segmentations_task_panel = CreatePngsFromSegmentationsTaskPanel()
        return self._create_pngs_from_segmentations_task_panel
    
    def calculate_scores_task_panel(self):
        if not self._calculate_scores_task_panel:
            self._calculate_scores_task_panel = CalculateScoresTaskPanel()
        return self._calculate_scores_task_panel
    
    def dicom2nifti_task_panel(self):
        if not self._dicom2nifti_task_panel:
            self._dicom2nifti_task_panel = Dicom2NiftiTaskPanel()
        return self._dicom2nifti_task_panel
    
    def create_dicom_summary_task_panel(self):
        if not self._create_dicom_summary_task_panel:
            self._create_dicom_summary_task_panel = CreateDicomSummaryTaskPanel()
        return self._create_dicom_summary_task_panel

    def select_slice_from_scans_task_panel(self):
        if not self._select_slice_from_scans_task_panel:
            self._select_slice_from_scans_task_panel = SelectSliceFromScansTaskPanel()
        return self._select_slice_from_scans_task_panel

    def total_segmentator_task_panel(self):
        if not self._total_segmentator_task_panel:
            self._total_segmentator_task_panel = TotalSegmentatorTaskPanel()
        return self._total_segmentator_task_panel
    
    def calculate_mask_statistics_task_panel(self):
        if not self._calculate_mask_statistics_task_panel:
            self._calculate_mask_statistics_task_panel = CalculateMaskStatisticsTaskPanel()
        return self._calculate_mask_statistics_task_panel
    
    def default_pipeline_panel(self):
        if not self._default_pipeline_panel:
            self._default_pipeline_panel = DefaultPipelinePanel()
        return self._default_pipeline_panel

    def default_docker_pipeline_panel(self):
        if not self._default_docker_pipeline_panel:
            self._default_docker_pipeline_panel = DefaultDockerPipelinePanel()
        return self._default_docker_pipeline_panel

    def boa_docker_pipeline_panel(self):
        if not self._boa_docker_pipeline_panel:
            self._boa_docker_pipeline_panel = BoaDockerPipelinePanel()
        return self._boa_docker_pipeline_panel
    
    def liver_analysis_pipeline_panel(self):
        if not self._liver_analysis_pipeline_panel:
            self._liver_analysis_pipeline_panel = LiverAnalysisPipelinePanel()
        return self._liver_analysis_pipeline_panel
    
    def slice_visualization(self):
        if not self._slice_visualization:
            self._slice_visualization = SliceVisualization()
        return self._slice_visualization

    # SETTERS

    def set_status(self, message):
        self.statusBar().showMessage(message)

    # EVENT HANDLERS

    def handle_rescale_dicom_images_task_action(self):
        self.main_panel().select_panel('rescaledicomimagestaskpanel')

    def handle_segment_muscle_fat_l3_tensorflow_task_action(self):
        self.main_panel().select_panel('segmentmusclefatl3tensorflowtaskpanel')

    def handle_segment_muscle_fat_pytorch_task_action(self):
        self.main_panel().select_panel('segmentmusclefatpytorchtaskpanel')

    def handle_create_pngs_from_segmentations_task_action(self):
        self.main_panel().select_panel('createpngsfromsegmentationstaskpanel')

    def handle_calculate_scores_task_action(self):
        self.main_panel().select_panel('calculatescorestaskpanel')

    def handle_dicom2nifti_task_action(self):
        self.main_panel().select_panel('dicom2niftitaskpanel')

    def handle_create_dicom_summary_task_action(self):
        self.main_panel().select_panel('createdicomsummarytaskpanel')

    def handle_select_slice_from_scans_task_action(self):
        self.main_panel().select_panel('selectslicefromscanstaskpanel')

    def handle_total_segmentator_task_action(self):
        self.main_panel().select_panel('totalsegmentatortaskpanel')

    def handle_calculate_mask_statistics_task_action(self):
        self.main_panel().select_panel('calculatemaskstatisticstaskpanel')

    def handle_default_pipeline_action(self):
        self.main_panel().select_panel('defaultpipelinepanel')

    def handle_default_docker_pipeline_action(self):
        self.main_panel().select_panel('defaultdockerpipelinepanel')

    def handle_boa_docker_pipeline_action(self):
        self.main_panel().select_panel('boadockerpipelinepanel')

    def handle_liver_analysis_pipeline_action(self):
        self.main_panel().select_panel('liveranalysispipelinepanel')

    def handle_slice_visualization_action(self):
        self.main_panel().select_panel('slicevisualization')

    def showEvent(self, event):
        return super().showEvent(event)

    def closeEvent(self, event):
        self.save_geometry_and_state()
        # Save inputs and parameters of relevant panels
        self.rescale_dicom_images_task_panel().save_inputs_and_parameters()
        self.segment_muscle_fat_l3_tensorflow_task_panel().save_inputs_and_parameters()
        self.segment_muscle_fat_pytorch_task_panel().save_inputs_and_parameters()
        self.create_pngs_from_segmentations_task_panel().save_inputs_and_parameters()
        self.calculate_scores_task_panel().save_inputs_and_parameters()
        self.dicom2nifti_task_panel().save_inputs_and_parameters()
        self.create_dicom_summary_task_panel().save_inputs_and_parameters()
        self.select_slice_from_scans_task_panel().save_inputs_and_parameters()
        self.total_segmentator_task_panel().save_inputs_and_parameters()
        self.calculate_mask_statistics_task_panel().save_inputs_and_parameters()
        self.default_pipeline_panel().save_inputs_and_parameters()
        self.default_docker_pipeline_panel().save_inputs_and_parameters()
        self.boa_docker_pipeline_panel().save_inputs_and_parameters()
        self.liver_analysis_pipeline_panel().save_inputs_and_parameters()
        self.slice_visualization().save_inputs_and_parameters()
        return super().closeEvent(event)

    def load_geometry_and_state(self):
        geometry = self.settings().get(constants.MOSAMATIC2_WINDOW_GEOMETRY_KEY)
        state = self.settings().get(constants.MOSAMATIC2_WINDOW_STATE_KEY)
        if isinstance(geometry, QByteArray) and self.restoreGeometry(geometry):
            if isinstance(state, QByteArray):
                self.restoreState(state)
            return True
        return False

    def save_geometry_and_state(self):
        self.settings().set(constants.MOSAMATIC2_WINDOW_GEOMETRY_KEY, self.saveGeometry())
        self.settings().set(constants.MOSAMATIC2_WINDOW_STATE_KEY, self.saveState())

    def set_default_size_and_position(self):
        self.resize(constants.MOSAMATIC2_WINDOW_W, constants.MOSAMATIC2_WINDOW_H)
        self.center_window()

    def center_window(self):
        screen = QGuiApplication.primaryScreen().geometry()
        x = (screen.width() - self.geometry().width()) / 2
        y = (screen.height() - self.geometry().height()) / 2
        self.move(int(x), int(y))


import os


class FileData:
    def __init__(self):
        self._path = None
        self._name = None
        self._object = None

    def path(self):
        return self._path
    
    def set_path(self, path):
        self._path = path

    def name(self):
        return os.path.split(self.path())[1]
    
    def object(self):
        return self._object
    
    def set_object(self, object):
        self._object = object
    
    def load(self):
        raise NotImplementedError()


import click

from mosamatic2.core.tasks import CalculateScoresTask


@click.command(help='Calculates body composition scores')
@click.option(
    '--images', 
    required=True, 
    type=click.Path(exists=True), 
    help='Directory with images',
)
@click.option(
    '--segmentations',
    required=True,
    type=click.Path(exists=True), 
    help='Directory with segmentations',
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory'
)
@click.option(
    '--file_type',
    default='npy',
    help='Options: [npy, tag]'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def calculatescores(images, segmentations, output, file_type, overwrite):
    """
    Calculates the following body composition metrics from the muscle and fat 
    images and segmentation files:

    (1) Muscle area (cm^2)
    (2) Mean muscle radiation attenuation (HU)
    (3) Subcutaneous fat area (cm^2)
    (4) Mean subcutaneous fat radiation attenuation (HU)
    (5) Visceral fat area (cm^2)
    (6) Mean visceral fat radiation attenuation (HU)
    
    Parameters
    ----------
    --images : str
        Directory with input L3 images

    --segmentations : str
        Directory with L3 muscle and fat segmenation files. Must be output of 
        "mosamatic2-cli segmentmusclefatl3tensorflow" OR a list of TAG files
        corresponding to the input images

    --output : str
        Path to output directory

    --file_type : str
        Type of segmentation file to use. Can be either "npy" or "tag"
    
    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    task = CalculateScoresTask(
        inputs={'images': images, 'segmentations': segmentations},
        params={'file_type': file_type},
        output=output,
        overwrite=overwrite,
    )
    task.run()


import os
import math
import tempfile
import shutil
import nibabel as nib
import numpy as np
from totalsegmentator.python_api import totalsegmentator
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import load_dicom

LOG = LogManager()

TOTAL_SEGMENTATOR_OUTPUT_DIR = os.path.join(tempfile.gettempdir(), 'total_segmentator_output')
TOTAL_SEGMENTATOR_TASK = 'total'
Z_DELTA_OFFSETS = {
    'vertebrae_L3': 0.333,
    'vertebrae_T4': 0.5,
}


class SelectSliceFromScansTask(Task):
    INPUTS = ['scans']
    PARAMS = ['vertebra']

    def __init__(self, inputs, params, output, overwrite):
        super(SelectSliceFromScansTask, self).__init__(inputs, params, output, overwrite)
        self._error_dir = os.path.split(self.output())[0]
        self._error_dir = os.path.join(self._error_dir, 'selectslicefromscanstask_errors')
        os.makedirs(self._error_dir, exist_ok=True)
        self._error_file = os.path.join(self._error_dir, 'errors.txt')
        with open(self._error_file, 'w') as f:
            f.write('Errors:\n\n')
        LOG.info(f'Error directory: {self._error_dir}')

    def write_error(self, message):
        LOG.error(message)
        with open(self._error_file, 'a') as f:
            f.write(message + '\n')

    def load_scan_dirs(self):
        scan_dirs = []
        for d in os.listdir(self.input('scans')):
            scan_dir = os.path.join(self.input('scans'), d)
            if os.path.isdir(scan_dir):
                scan_dirs.append(scan_dir)
        return scan_dirs
    
    def extract_masks(self, scan_dir):
        os.makedirs(TOTAL_SEGMENTATOR_OUTPUT_DIR, exist_ok=True)
        totalsegmentator(input=scan_dir, output=TOTAL_SEGMENTATOR_OUTPUT_DIR, fast=True)
        if not os.path.isfile(os.path.join(TOTAL_SEGMENTATOR_OUTPUT_DIR, 'vertebrae_L3.nii.gz')):
            raise Exception(f'{scan_dir}: vertebrae_L3.nii.gz does not exist')
        # os.system(f'TotalSegmentator -i {scan_dir} -o {TOTAL_SEGMENTATOR_OUTPUT_DIR} --fast')

    def delete_total_segmentator_output(self):
        if os.path.exists(TOTAL_SEGMENTATOR_OUTPUT_DIR):
            shutil.rmtree(TOTAL_SEGMENTATOR_OUTPUT_DIR)

    def get_z_delta_offset_for_mask(self, mask_name):
        if mask_name not in Z_DELTA_OFFSETS.keys():
            return None
        return Z_DELTA_OFFSETS[mask_name]

    def find_slice(self, scan_dir, vertebra):
        if vertebra == 'L3':
            vertebral_level = 'vertebrae_L3'
        elif vertebra == 'T4':
            vertebral_level = 'vertebrae_T4'
        else:
            self.write_error(f'{scan_dir}: Unknown vertbra {vertebra}. Options are "L3" and "T4"')
            return None
        # Find Z-positions DICOM images
        z_positions = {}
        for f in os.listdir(scan_dir):
            f_path = os.path.join(scan_dir, f)
            try:
                p = load_dicom(f_path, stop_before_pixels=True)
                if p is not None and hasattr(p, "ImagePositionPatient"):
                    z_positions[p.ImagePositionPatient[2]] = f_path
            except Exception as e:
                self.write_error(f"{scan_dir}: Failed to load DICOM {f_path}: {e}")
                break
        if not z_positions:
            self.write_error(f"{scan_dir}: No valid DICOM z-positions found.")
            return None
        # Find Z-position L3 image
        mask_file = os.path.join(TOTAL_SEGMENTATOR_OUTPUT_DIR, f'{vertebral_level}.nii.gz')
        if not os.path.exists(mask_file):
            self.write_error(f"{scan_dir}: Mask file not found: {mask_file}")
            return None
        try:
            mask_obj = nib.load(mask_file)
            mask = mask_obj.get_fdata()
            affine_transform = mask_obj.affine
        except Exception as e:
            self.write_error(f"{scan_dir}: Failed to load mask {mask_file}: {e}")
            return None
        indexes = np.array(np.where(mask == 1))
        if indexes.size == 0:
            self.write_error(f"{scan_dir}: No voxels found in mask {mask_file} for {vertebral_level}")
            return None        
        try:
            index_min = indexes.min(axis=1)
            index_max = indexes.max(axis=1)
        except ValueError as e:
            self.write_error(f"{scan_dir}: Invalid indexes array for {vertebral_level}: {e}")
            return None
        world_min = nib.affines.apply_affine(affine_transform, index_min)
        world_max = nib.affines.apply_affine(affine_transform, index_max)
        z_direction = affine_transform[:3, 2][2]
        if z_direction == 0:
            self.write_error(f"{scan_dir}: Affine z-direction is zero.")
            return None
        z_sign = math.copysign(1, z_direction)
        z_delta_offset = self.get_z_delta_offset_for_mask(vertebral_level)
        if z_delta_offset is None:
            return None
        z_delta = 0.333 * abs(world_max[2] - world_min[2]) # This needs to be vertebra-specific perhaps
        z_l3 = world_max[2] - z_sign * z_delta
        # Find closest L3 image in DICOM set
        positions = sorted(z_positions.keys())
        closest_file = None
        for z1, z2 in zip(positions[:-1], positions[1:]):
            if min(z1, z2) <= z_l3 <= max(z1, z2):
                closest_z = min(z_positions.keys(), key=lambda z: abs(z - z_l3))
                closest_file = z_positions[closest_z]
                LOG.info(f'Closest image: {closest_file}')
                break
        if closest_file is None:
            self.write_error(f"{scan_dir}: No matching slice found.")
        return closest_file

    def run(self):
        scan_dirs = self.load_scan_dirs()
        vertebra = self.param('vertebra')
        nr_steps = len(scan_dirs)
        for step in range(nr_steps):
            scan_dir = scan_dirs[step]
            scan_name = os.path.split(scan_dir)[1]
            errors = False
            LOG.info(f'Processing {scan_dir}...')
            try:
                self.extract_masks(scan_dir)
            except Exception as e:
                self.write_error(f'{scan_dir}: Could not extract masks [{str(e)}]. Skipping scan...')
                errors = True
            if not errors:
                file_path = self.find_slice(scan_dir, vertebra)
                if file_path is not None:
                    extension = '' if file_path.endswith('.dcm') else '.dcm'
                    target_file_path = os.path.join(self.output(), vertebra + '_' + scan_name + extension)
                    shutil.copyfile(file_path, target_file_path)
                else:
                    self.write_error(f'{scan_dir}: Could not find slice for vertebral level: {vertebra}')
                    errors = True
                self.delete_total_segmentator_output()
            if errors:
                LOG.info(f'Copying problematic scan {scan_dir} to error directory: {self._error_dir}')
                scan_error_dir = os.path.join(self._error_dir, scan_name)
                shutil.copytree(scan_dir, scan_error_dir)
            self.set_progress(step, nr_steps)